/* Fonts */

@font-face {
  font-family: 'manrope__bold';
  src: url('../fonts/Manrope-Bold.ttf');
  font-weight: normal;
  font-style: normal;
  font-display: swap;
}

@font-face {
  font-family: 'inter__regular';
  src: url('../fonts/Inter-Regular.ttf');
  font-weight: normal;
  font-style: normal;
  font-display: swap;
}

@font-face {
  font-family: 'inter__bold';
  src: url('../fonts/Inter-Bold.ttf');
  font-weight: normal;
  font-style: normal;
  font-display: swap;
}


/* can add more fonts from autonolas repo if more fonts are needed */

:root {
  --md-primary-fg-color: #7200D6;
  --md-primary-fg-color--light: rgb(0 217 72 / 10%);
  --md-primary-fg-color--dark: #086228;
  --md-typeset-a-color: var(--md-primary-fg-color--dark);

  --md-extra-color-white: #FFFFFF;
  --md-extra-color-black: #000000;
  --md-extra-color-gray-1: #585858;
  --md-extra-color-gray-2: #f8f8f7;
  --md-extra-color-gray-3: #000000;
  --md-body-font: "inter__regular", "Roboto", "Helvetica Neue", Helvetica, Arial, sans-serif;
  --md-heading-font: "manrope__bold", var(--body-font);
}

.md-content__inner h1 {
  font-family: var(--md-heading-font);
}

.md-nav--primary .md-nav__title {
  display: none;
}

/* header & tabs */
.md-header {
  color: var(--md-extra-color-white);
}

.md-header__title {
  margin-left: 0;
}

/* hiding icon at top */
.md-header-nav .md-flex__cell.md-flex__cell--shrink {
  display: none;
}

.md-header-nav .md-header-nav__title .md-header-nav__topic {
  font-size: 24px;
}

.md-header-nav .md-header-nav__title .md-header-nav__topic:nth-child(2) {
  letter-spacing: normal;
}

.md-header-nav .md-flex .md-flex__cell:last-child {
  display: none;
}

.md-tabs .md-tabs__list>.md-tabs__item a {
  color: var(--md-extra-color-black);
}

.md-tabs .md-tabs__list>.md-tabs__item {
  padding-left: 1rem;
  padding-right: 1rem;
}

.md-tabs__link--active,
.md-tabs__link:hover {
  text-decoration: underline;
}

/* left sidebar */
.md-sidebar .md-nav .md-nav__title {
  color: var(--md-extra-color-black);
  font-size: 18px;
}


.md-sidebar .md-nav__link.md-nav__link {
  color: var(--md-extra-color-gray-1);
}

.md-sidebar .md-nav__link.md-nav__link--active,
.md-sidebar .md-nav__link:active {
  text-decoration: underline;
  color: var(--md-extra-color-black);
}

.md-sidebar .md-nav__link:focus,
.md-nav__link:hover {
  color: var(--md-primary-fg-color--dark);
}

/* right sidebar */

/* ------ content ------ */
.md-content__inner h1 {
  font-weight: bold;
  color: var(--md-extra-color-black);
}

.md-typeset a {
  color: var(--md-typeset-a-color);
  font-weight: bold;
}

.md-typeset a:active,
.md-typeset a:hover {
  color: var(--md-typeset-a-color);
}

/* ------ footer ------ */
.md-footer {
  background-color: var(--md-primary-fg-color);
}

.md-footer .md-footer-nav,
.md-footer .md-footer-nav__direction {
  color: var(--md-extra-color-black);
  font-family: var(--md-body-font);
}

.md-footer .md-footer-nav__title {
  text-transform: uppercase;
}

.md-footer .md-footer-nav__direction {
  text-transform: none;
}

.md-footer .md-footer-meta {
  display: none;
}

/* ------ common ------ */
body {
  font-family: var(--md-body-font);
}

pre {
  background-color: #f8f8f7;
  background-color: var(--md-extra-color-gray-2);
}

code {
  background: var(--md-primary-fg-color);
}

.md-header,
.md-tabs,
.md-footer-meta,
.md-footer-nav,
.md-footer-nav__inner {
  background-color: var(--md-primary-fg-color);
}

.md-nav__title {
  color: var(--md-primary-fg-color--dark);
}

/* Needed so that Mermaid UML diagrams don't end up being massively tall */
svg {
  height: auto;
}
