# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ipfs.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\nipfs.proto\x12\x16\x61\x65\x61.valory.ipfs.v0_1_0"\xb2\x07\n\x0bIpfsMessage\x12G\n\x05\x65rror\x18\x05 \x01(\x0b\x32\x36.aea.valory.ipfs.v0_1_0.IpfsMessage.Error_PerformativeH\x00\x12G\n\x05\x66iles\x18\x06 \x01(\x0b\x32\x36.aea.valory.ipfs.v0_1_0.IpfsMessage.Files_PerformativeH\x00\x12O\n\tget_files\x18\x07 \x01(\x0b\x32:.aea.valory.ipfs.v0_1_0.IpfsMessage.Get_Files_PerformativeH\x00\x12O\n\tipfs_hash\x18\x08 \x01(\x0b\x32:.aea.valory.ipfs.v0_1_0.IpfsMessage.Ipfs_Hash_PerformativeH\x00\x12S\n\x0bstore_files\x18\t \x01(\x0b\x32<.aea.valory.ipfs.v0_1_0.IpfsMessage.Store_Files_PerformativeH\x00\x1a\xc9\x01\n\x18Store_Files_Performative\x12V\n\x05\x66iles\x18\x01 \x03(\x0b\x32G.aea.valory.ipfs.v0_1_0.IpfsMessage.Store_Files_Performative.FilesEntry\x12\x0f\n\x07timeout\x18\x02 \x01(\x01\x12\x16\n\x0etimeout_is_set\x18\x03 \x01(\x08\x1a,\n\nFilesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a+\n\x16Ipfs_Hash_Performative\x12\x11\n\tipfs_hash\x18\x01 \x01(\t\x1aT\n\x16Get_Files_Performative\x12\x11\n\tipfs_hash\x18\x01 \x01(\t\x12\x0f\n\x07timeout\x18\x02 \x01(\x01\x12\x16\n\x0etimeout_is_set\x18\x03 \x01(\x08\x1a\x94\x01\n\x12\x46iles_Performative\x12P\n\x05\x66iles\x18\x01 \x03(\x0b\x32\x41.aea.valory.ipfs.v0_1_0.IpfsMessage.Files_Performative.FilesEntry\x1a,\n\nFilesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a$\n\x12\x45rror_Performative\x12\x0e\n\x06reason\x18\x01 \x01(\tB\x0e\n\x0cperformativeb\x06proto3'
)


_IPFSMESSAGE = DESCRIPTOR.message_types_by_name["IpfsMessage"]
_IPFSMESSAGE_STORE_FILES_PERFORMATIVE = _IPFSMESSAGE.nested_types_by_name[
    "Store_Files_Performative"
]
_IPFSMESSAGE_STORE_FILES_PERFORMATIVE_FILESENTRY = (
    _IPFSMESSAGE_STORE_FILES_PERFORMATIVE.nested_types_by_name["FilesEntry"]
)
_IPFSMESSAGE_IPFS_HASH_PERFORMATIVE = _IPFSMESSAGE.nested_types_by_name[
    "Ipfs_Hash_Performative"
]
_IPFSMESSAGE_GET_FILES_PERFORMATIVE = _IPFSMESSAGE.nested_types_by_name[
    "Get_Files_Performative"
]
_IPFSMESSAGE_FILES_PERFORMATIVE = _IPFSMESSAGE.nested_types_by_name[
    "Files_Performative"
]
_IPFSMESSAGE_FILES_PERFORMATIVE_FILESENTRY = (
    _IPFSMESSAGE_FILES_PERFORMATIVE.nested_types_by_name["FilesEntry"]
)
_IPFSMESSAGE_ERROR_PERFORMATIVE = _IPFSMESSAGE.nested_types_by_name[
    "Error_Performative"
]
IpfsMessage = _reflection.GeneratedProtocolMessageType(
    "IpfsMessage",
    (_message.Message,),
    {
        "Store_Files_Performative": _reflection.GeneratedProtocolMessageType(
            "Store_Files_Performative",
            (_message.Message,),
            {
                "FilesEntry": _reflection.GeneratedProtocolMessageType(
                    "FilesEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _IPFSMESSAGE_STORE_FILES_PERFORMATIVE_FILESENTRY,
                        "__module__": "ipfs_pb2"
                        # @@protoc_insertion_point(class_scope:aea.valory.ipfs.v0_1_0.IpfsMessage.Store_Files_Performative.FilesEntry)
                    },
                ),
                "DESCRIPTOR": _IPFSMESSAGE_STORE_FILES_PERFORMATIVE,
                "__module__": "ipfs_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ipfs.v0_1_0.IpfsMessage.Store_Files_Performative)
            },
        ),
        "Ipfs_Hash_Performative": _reflection.GeneratedProtocolMessageType(
            "Ipfs_Hash_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _IPFSMESSAGE_IPFS_HASH_PERFORMATIVE,
                "__module__": "ipfs_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ipfs.v0_1_0.IpfsMessage.Ipfs_Hash_Performative)
            },
        ),
        "Get_Files_Performative": _reflection.GeneratedProtocolMessageType(
            "Get_Files_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _IPFSMESSAGE_GET_FILES_PERFORMATIVE,
                "__module__": "ipfs_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ipfs.v0_1_0.IpfsMessage.Get_Files_Performative)
            },
        ),
        "Files_Performative": _reflection.GeneratedProtocolMessageType(
            "Files_Performative",
            (_message.Message,),
            {
                "FilesEntry": _reflection.GeneratedProtocolMessageType(
                    "FilesEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _IPFSMESSAGE_FILES_PERFORMATIVE_FILESENTRY,
                        "__module__": "ipfs_pb2"
                        # @@protoc_insertion_point(class_scope:aea.valory.ipfs.v0_1_0.IpfsMessage.Files_Performative.FilesEntry)
                    },
                ),
                "DESCRIPTOR": _IPFSMESSAGE_FILES_PERFORMATIVE,
                "__module__": "ipfs_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ipfs.v0_1_0.IpfsMessage.Files_Performative)
            },
        ),
        "Error_Performative": _reflection.GeneratedProtocolMessageType(
            "Error_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _IPFSMESSAGE_ERROR_PERFORMATIVE,
                "__module__": "ipfs_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ipfs.v0_1_0.IpfsMessage.Error_Performative)
            },
        ),
        "DESCRIPTOR": _IPFSMESSAGE,
        "__module__": "ipfs_pb2"
        # @@protoc_insertion_point(class_scope:aea.valory.ipfs.v0_1_0.IpfsMessage)
    },
)
_sym_db.RegisterMessage(IpfsMessage)
_sym_db.RegisterMessage(IpfsMessage.Store_Files_Performative)
_sym_db.RegisterMessage(IpfsMessage.Store_Files_Performative.FilesEntry)
_sym_db.RegisterMessage(IpfsMessage.Ipfs_Hash_Performative)
_sym_db.RegisterMessage(IpfsMessage.Get_Files_Performative)
_sym_db.RegisterMessage(IpfsMessage.Files_Performative)
_sym_db.RegisterMessage(IpfsMessage.Files_Performative.FilesEntry)
_sym_db.RegisterMessage(IpfsMessage.Error_Performative)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _IPFSMESSAGE_STORE_FILES_PERFORMATIVE_FILESENTRY._options = None
    _IPFSMESSAGE_STORE_FILES_PERFORMATIVE_FILESENTRY._serialized_options = b"8\001"
    _IPFSMESSAGE_FILES_PERFORMATIVE_FILESENTRY._options = None
    _IPFSMESSAGE_FILES_PERFORMATIVE_FILESENTRY._serialized_options = b"8\001"
    _IPFSMESSAGE._serialized_start = 39
    _IPFSMESSAGE._serialized_end = 985
    _IPFSMESSAGE_STORE_FILES_PERFORMATIVE._serialized_start = 448
    _IPFSMESSAGE_STORE_FILES_PERFORMATIVE._serialized_end = 649
    _IPFSMESSAGE_STORE_FILES_PERFORMATIVE_FILESENTRY._serialized_start = 605
    _IPFSMESSAGE_STORE_FILES_PERFORMATIVE_FILESENTRY._serialized_end = 649
    _IPFSMESSAGE_IPFS_HASH_PERFORMATIVE._serialized_start = 651
    _IPFSMESSAGE_IPFS_HASH_PERFORMATIVE._serialized_end = 694
    _IPFSMESSAGE_GET_FILES_PERFORMATIVE._serialized_start = 696
    _IPFSMESSAGE_GET_FILES_PERFORMATIVE._serialized_end = 780
    _IPFSMESSAGE_FILES_PERFORMATIVE._serialized_start = 783
    _IPFSMESSAGE_FILES_PERFORMATIVE._serialized_end = 931
    _IPFSMESSAGE_FILES_PERFORMATIVE_FILESENTRY._serialized_start = 605
    _IPFSMESSAGE_FILES_PERFORMATIVE_FILESENTRY._serialized_end = 649
    _IPFSMESSAGE_ERROR_PERFORMATIVE._serialized_start = 933
    _IPFSMESSAGE_ERROR_PERFORMATIVE._serialized_end = 969
# @@protoc_insertion_point(module_scope)
